implementation module Corba

import StdEnv, StdDebug
import corba_wrappers

import RWSDebug, Debug

from StdString import String;

////////////////////////////////
//       UTILITIES
////////////////////////////////

was_exception :: CorbaException -> Bool
was_exception NoException 	= False
was_exception _ 			= True

:: ResultOrException a b = Result a | Exception b

// If there was an exception, the result is CorbaVoid
// Luckily, lazy evaluation prevents the system from trying to decode
// it into another CORBA type...
corba_result_create :: .a CorbaException -> ResultOrException .a CorbaException
corba_result_create result exception
	| was_exception exception
	= Exception exception
	| otherwise
	= Result result


:: MayBe a = Just a | Nothing

just :: (MayBe u:a) -> u:a
just (Just a) 	= a

is_just :: (MayBe .a) -> Bool
is_just (Just a) = True
is_just (Nothing) = False

toStr :: !.a -> String
toStr a = foldl (+++) "" (debugShowWithOptions [] a)

map_with_env :: (a *env -> (b, *env)) [a] *env -> ([b], *env)
map_with_env f [] env = ([], env)
map_with_env f [x:xs] env
	# (x2, env) = f x env
	# (x3, env)	= map_with_env f xs env
	= ([x2:x3], env)

map_maybe_with_env :: (a *env -> (MayBe b, *env)) [a] *env -> (MayBe [b], *env)
map_maybe_with_env f l env
	# (l, env) = map_with_env f l env
	| and (map is_just l)
	= (Just (map just l), env)
	| otherwise
	= (Nothing, env)

////////////////////////////////
//         MONADS
////////////////////////////////

:: IO a :== *World -> (a, *World)

(>>=) f g :== f `bind` g

prodM :: [IO a] -> IO [a]

prodM [] = return []
prodM [m:ms] =
	m >>= \x ->
	prodM ms >>= \xs ->
	return [x:xs]

void :: (*World -> *World) -> IO Void
void f = \w -> (Void, f w)

////////////////////////////////
//  HANDLING CORBA SEQUENCES
////////////////////////////////

// Higher-order function for creating CORBA Sequences from a Clean list

:: SeqCreateFct seq_type			:==	Int *World -> (seq_type, *World)
:: SeqSetFct seq_type member_type	:==	seq_type Int member_type *World -> *World

corba_seq_encode :: (SeqCreateFct seq_type)
					(SeqSetFct seq_type member_type)
					[member_type] 
					*World
					-> (seq_type, *World)
corba_seq_encode create_fun set_fun l env
	# (seq, env)	=	create_fun (length l) env
	# (_, env)		=   (prodM [ void (set_fun seq i v) \\ i<-[0..] & v <- l ]) env
	= (seq, env)

:: SeqGetFct seq_type member_type 	:==	seq_type Int -> IO member_type

// Higher-order function for decoding CORBA Sequences into a Clean list

corba_seq_decode :: (SeqGetFct seq_type member_type) seq_type Int
					-> IO [member_type]
corba_seq_decode get_fun seq len
	= prodM [ (get_fun seq i) \\ i <- [0..len-1] ]

////////////////////////////////
//       CStringArray
////////////////////////////////

c_string_array_encode :: [String] *World -> (CStringArrayPtr, Int, *World)
c_string_array_encode l env
	# (arr, env) = corba_seq_encode c_string_array_create set_member l env
	= (arr, length l, env)
where
	set_member arr index value env = c_string_array_set arr index value env

c_string_array_decode :: CStringArrayPtr Int *World -> ([String], *World)
c_string_array_decode arr len env = corba_seq_decode decode_member arr len env
where
	decode_member arr index env
	# (value, env) = c_string_array_get arr index env
	= (value, env)


////////////////////////////////
//        TYPECODE
////////////////////////////////

// HIGH LEVEL INTERFACE

:: TypeCode = 	Tk_Null | 
				Tk_Void | 
				Tk_Short | 
				Tk_Long | 
				Tk_UShort |
				Tk_ULong | 
				Tk_Float | 
				Tk_Double | 
				Tk_Boolean | 
				Tk_Char |
				Tk_Octet | 
				Tk_Any | 
				Tk_String Int | 
				Tk_TypeCode | 
				Tk_ObjRef String String | 
				Tk_Struct String String [(String, TypeCode)] | 
			  	Tk_Sequence Int TypeCode |
			  	Tk_Union String String TypeCode [(String, Any, TypeCode)] |
				Tk_Enum String String [String] |
				Tk_Alias String String TypeCode |
				Tk_Array Int TypeCode |
				Tk_Exception String String [(String, TypeCode)] |
				Tk_Recursive String

instance == TypeCode where
	(==) tc1 tc2	=	True

// These functions could be implemented more elegantly using overloaded
// functions, but Clean does not allow overloading with two class
// variables (sequence type, member type).

// The 'repoids' parameter is used for handling recursive typecodes
corba_struct_member_seq_encode_with :: [(String, TypeCode)] [String] *World -> (StructMemberSeqPtr, *World)
corba_struct_member_seq_encode_with members repoids env
	= corba_seq_encode (corba_struct_member_seq_create) (add_member) members env
	where
		add_member :: StructMemberSeqPtr Int (String, TypeCode) *World -> *World
		add_member seq index (name, type) env
		# (tcPtr, env) = corba_tc_encode_with type repoids env
		= corba_struct_member_seq_fill seq index name tcPtr env

corba_union_member_seq_encode_with :: [(String, Any, TypeCode)] [String] *World -> (UnionMemberSeqPtr, *World)
corba_union_member_seq_encode_with members repoids env
	= corba_seq_encode (corba_union_member_seq_create) (add_member) members env
	where
		add_member :: UnionMemberSeqPtr Int (String, Any, TypeCode) *World -> *World
		add_member seq index (name, label, type) env
		# (labelPtr, env)	=	corba_any_encode label env
		# (typePtr, env)	=	corba_tc_encode_with type repoids env
		# env = corba_union_member_seq_fill seq index name labelPtr typePtr env
		= env

corba_enum_member_seq_encode :: [String] *World -> (EnumMemberSeqPtr, *World)
corba_enum_member_seq_encode members  env
	= corba_seq_encode (corba_enum_member_seq_create) (add_member) members env
	where
		add_member :: EnumMemberSeqPtr Int String *World -> *World
		add_member seq index name env = corba_enum_member_seq_fill seq index name env

//
// Handling recursive typecodes:
//   The 'encode' and 'decode' functions deals with recursive typecodes 
// automatically, by inserting Tk_Recursive typecodes where appropriate.
// This means the call is free to pass infinite typecodes such as
// tc = (Tk_Struct "x" "x" [("bla", (Tk_Sequence 0 tc))])
//

corba_tc_encode :: TypeCode *World -> (TypeCodePtr, *World)
corba_tc_encode tc env = corba_tc_encode_with tc [] env

// The second argument contains the repoids of IDL structures and unions
// which contain this typecode.
corba_tc_encode_with :: TypeCode [String] *World -> (TypeCodePtr, *World)

corba_tc_encode_with (Tk_Null) repoids env 	= corba_tc_create_basic TK_NULL env
corba_tc_encode_with (Tk_Void) repoids env		= corba_tc_create_basic TK_VOID env
corba_tc_encode_with (Tk_Short) repoids env		= corba_tc_create_basic TK_SHORT env
corba_tc_encode_with (Tk_Long) repoids env		= corba_tc_create_basic TK_LONG env
corba_tc_encode_with (Tk_UShort) repoids env		= corba_tc_create_basic TK_USHORT env
corba_tc_encode_with (Tk_ULong) repoids env		= corba_tc_create_basic TK_ULONG env
corba_tc_encode_with (Tk_Float) repoids env		= corba_tc_create_basic TK_FLOAT env
corba_tc_encode_with (Tk_Double) repoids env		= corba_tc_create_basic TK_DOUBLE env
corba_tc_encode_with (Tk_Boolean) repoids env	= corba_tc_create_basic TK_BOOLEAN env
corba_tc_encode_with (Tk_Char) repoids env		= corba_tc_create_basic TK_CHAR env
corba_tc_encode_with (Tk_Octet) repoids env		= corba_tc_create_basic TK_OCTET env
corba_tc_encode_with (Tk_Any) repoids env		= corba_tc_create_basic TK_ANY env
corba_tc_encode_with (Tk_String bound) repoids env 	= corba_tc_create_string bound env
corba_tc_encode_with (Tk_TypeCode) repoids env	= corba_tc_create_basic TK_TYPECODE env
corba_tc_encode_with (Tk_ObjRef repoid name) repoids env	= corba_tc_create_objref repoid name env

corba_tc_encode_with (Tk_Struct repoid name members) repoids env
	| isMember repoid repoids
	= corba_tc_create_recursive repoid env
	| otherwise
	# (seq, env) 	=	corba_struct_member_seq_encode_with members [repoid:repoids] env
	= corba_tc_create_struct repoid name seq env

corba_tc_encode_with (Tk_Sequence bound member_type) repoids env
	# (tc, env) 	=	corba_tc_encode_with member_type repoids env
	= corba_tc_create_sequence bound tc env

corba_tc_encode_with (Tk_Union repoid name label_type members) repoids env
	| isMember repoid repoids
	= corba_tc_create_recursive repoid env
	| otherwise
	# (label_tc, env) =	corba_tc_encode_with label_type repoids env
	# (seq, env) 	=	corba_union_member_seq_encode_with members [repoid:repoids] env
	= corba_tc_create_union repoid name label_tc seq env

corba_tc_encode_with (Tk_Enum repoid name members) repoids env
	# (seq, env) 	=	corba_enum_member_seq_encode members env
	= corba_tc_create_enum repoid name seq env

corba_tc_encode_with (Tk_Alias repoid name content_type) repoids env
	# (content_tc, env) =	corba_tc_encode_with content_type repoids env
	= corba_tc_create_alias repoid name content_tc env

corba_tc_encode_with (Tk_Array length member_type) repoids env
	# (member_tc, env) =	corba_tc_encode_with member_type repoids env
	= corba_tc_create_array member_tc length env

corba_tc_encode_with (Tk_Exception repoid name members) repoids env
	# (seq, env) 	=	corba_struct_member_seq_encode_with members repoids env
	= corba_tc_create_exception repoid name seq env

corba_tc_encode_with (Tk_Recursive repoid) repoids env
	= corba_tc_create_recursive repoid env


corba_tc_decode :: TypeCodePtr *World -> (TypeCode, *World)
corba_tc_decode tc env
	# (kind, env)	=	corba_tc_get_kind tc env
	= decode_kind tc kind env
where
	decode_kind :: TypeCodePtr Int *World -> (TypeCode, *World)
	decode_kind tc TK_NULL env 		= (Tk_Null, env)
	decode_kind tc TK_VOID 	 env	= (Tk_Void, env)
	decode_kind tc TK_SHORT  env	= (Tk_Short, env)
	decode_kind tc TK_LONG 	 env	= (Tk_Long, env)
	decode_kind tc TK_USHORT  env	= (Tk_UShort, env)
	decode_kind tc TK_ULONG  env	= (Tk_ULong, env)
	decode_kind tc TK_FLOAT  env	= (Tk_Float, env)
	decode_kind tc TK_DOUBLE  env	= (Tk_Double, env)
	decode_kind tc TK_BOOLEAN  env	= (Tk_Boolean, env)
	decode_kind tc TK_CHAR 	 env	= (Tk_Char, env)
	decode_kind tc TK_OCTET  env	= (Tk_Octet, env)
	decode_kind tc TK_ANY	 env	= (Tk_Any, env)

	decode_kind tc TK_STRING env
	# (length, env)				= corba_tc_get_length tc env
	= (Tk_String length, env)

	decode_kind tc TK_TYPECODE env	= (Tk_TypeCode, env)


	decode_kind tc TK_OBJREF env
	# (repoid, env) = corba_tc_get_id tc env
	# (name, env) = corba_tc_get_name tc env
	= (Tk_ObjRef repoid name, env)

	decode_kind tc TK_STRUCT env = decode_struct tc env

	//
	// This function never terminates if the sequence is recursive.
	// To avoid this, we check if the content type is a struct or union,
	// and return a Tk_Recursive type code instead of decoding the content
	// type.
	//

	decode_kind tc TK_SEQUENCE env
		# (length, env) 		= corba_tc_get_length tc env
		# (contentTypePtr, env) = corba_tc_get_content_type tc env
		# (content_kind, env)	= corba_tc_get_kind contentTypePtr env
		| content_kind == TK_STRUCT || content_kind == TK_UNION
		# (content_repoid, env) = corba_tc_get_id contentTypePtr env
		= (Tk_Sequence length (Tk_Recursive content_repoid), env)
		| otherwise
		# (content_type, env)	=	corba_tc_decode contentTypePtr env
		= (Tk_Sequence length content_type, env)

	decode_kind tc TK_UNION env = decode_union tc env

	decode_kind tc TK_ENUM env = decode_enum tc env

	decode_kind tc TK_ALIAS env
	# (repoid, env)			= corba_tc_get_id tc env
	# (name, env)			= corba_tc_get_name tc env
	# (contentTypePtr, env)	= corba_tc_get_content_type tc env
	# (content_type, env)	= corba_tc_decode contentTypePtr env
	= (Tk_Alias repoid name content_type, env)

	decode_kind tc TK_ARRAY env
	# (contentTypePtr, env)	= corba_tc_get_content_type tc env
	# (length, env)			= corba_tc_get_length tc env
	# (content_type, env)		= corba_tc_decode contentTypePtr env
	= (Tk_Array length content_type, env)

	decode_kind tc TK_EXCEPT env = decode_exception tc env

	decode_kind tc _ _			= abort "unknown TCKind in decode_kind."

	// TypeCodes for structs are created using StructMemberSeq-s, and
	// queried using functions of the TypeCode object... :(
	decode_struct :: TypeCodePtr *World -> (TypeCode, *World)
	decode_struct tc env
	# (repoid, env)			= corba_tc_get_id tc env
	# (name, env)			= corba_tc_get_name tc env
	# (member_count, env) 	= corba_tc_get_member_count tc env
	# (members, env)		= corba_seq_decode get_member tc member_count env
	= (Tk_Struct repoid name members, env)
	where
		get_member :: TypeCodePtr Int *World -> ((String, TypeCode), *World)
		get_member tc index env
		# (member_name, env)	= corba_tc_get_member_name tc index env
		# (memberTypePtr, env)	= corba_tc_get_member_type tc index env
		# (member_type, env)		= corba_tc_decode memberTypePtr env
		= ((member_name, member_type), env)

	// Same as decode_struct
	decode_exception :: TypeCodePtr *World -> (TypeCode, *World)
	decode_exception tc env
	# (repoid, env)			= corba_tc_get_id tc env
	# (name, env)			= corba_tc_get_name tc env
	# (member_count, env)	= corba_tc_get_member_count tc env
	# (members, env)			= corba_seq_decode decode_member tc member_count env
	= (Tk_Exception repoid name members, env)
	where
		decode_member :: TypeCodePtr Int *World -> ((String, TypeCode), *World)
		decode_member tc index env
		# (member_name, env)	= corba_tc_get_member_name tc index env
		# (typePtr, env) 		= corba_tc_get_member_type tc index env
		# (member_type, env)		= corba_tc_decode typePtr env
		= ((member_name, member_type), env)

	decode_union :: TypeCodePtr *World -> (TypeCode, *World)
	decode_union tc env
	# (repoid, env)			= corba_tc_get_id tc env
	# (name, env)			= corba_tc_get_name tc env
	# (member_count, env)	= corba_tc_get_member_count tc env
	# (discTypePtr, env)	= corba_tc_get_discriminator_type tc env
	# (discType, env)			= corba_tc_decode discTypePtr env
	# (members, env)			= corba_seq_decode decode_member tc member_count env
	= (Tk_Union repoid name discType members, env)
	where
		decode_member :: TypeCodePtr Int *World -> ((String, Any, TypeCode), *World)
		decode_member tc index env
		# (member_name, env)	= corba_tc_get_member_name tc index env
		# (labelPtr, env)	 	= corba_tc_get_member_label tc index env
		# (typePtr, env)	 	= corba_tc_get_member_type tc index env
		# (member_label, env)		= corba_any_decode labelPtr env
		# (member_type, env)		= corba_tc_decode typePtr env
		= ((member_name, member_label, member_type), env)

	decode_enum :: TypeCodePtr *World -> (TypeCode, *World)
	decode_enum tc env
	# (repoid, env)			= corba_tc_get_id tc env
	# (name, env)			= corba_tc_get_name tc env
	# (member_count, env)	= corba_tc_get_member_count tc env
	# (members, env)			= corba_seq_decode get_member tc member_count env
	= (Tk_Enum repoid name members, env)
	where
		get_member :: TypeCodePtr Int *World -> (String, *World)
		get_member tc index env
		= corba_tc_get_member_name tc index env

////////////////////////////////
//          CORBA::Any
////////////////////////////////

// HIGH LEVEL INTERFACE

:: Any = 	CorbaNull | 
			CorbaVoid | 
			CorbaShort Int | 
			CorbaLong Int |
		   	CorbaUShort Int | 
			CorbaULong Int | 
			CorbaFloat Real |
			CorbaDouble Real |
			CorbaBoolean Bool |
			CorbaChar Char |
			CorbaOctet Int |
			CorbaString String |
			CorbaAny Any |
			CorbaTypeCode TypeCode |
			CorbaObjRef CORBA_Object |
			CorbaStruct TypeCode [(String, Any)] |  // tc [(name, type)]
			CorbaSequence TypeCode [Any] |			// tc [members]
			CorbaUnion TypeCode Any Any	|			// tc discriminator value
			CorbaEnum TypeCode String |				// tc value
			CorbaArray TypeCode [Any] |				// tc [members]
			CorbaException TypeCode [(String, Any)]
//
// If there is no active element in a CorbaUnion, the second argument
// is CorbaNull.
//

corba_nv_pair_seq_encode :: [(String, Any)] *World -> (NameValuePairSeq, *World)
corba_nv_pair_seq_encode members env
	= corba_seq_encode corba_nv_pair_seq_create add_member members env
	where
		add_member :: NameValuePairSeq Int (String, Any) *World -> *World
		add_member seq index (name, value) env
		# (valuePtr, env)		=	corba_any_encode value env
		# env = corba_nv_pair_seq_fill seq index name valuePtr env
		= env

corba_nv_pair_seq_decode :: NameValuePairSeq *World -> ([(String, Any)], *World)
corba_nv_pair_seq_decode seq env
	# (len, env) = corba_nv_pair_seq_length seq env
	= corba_seq_decode decode_member seq len env
	where
		decode_member :: NameValuePairSeq Int *World -> ((String, Any), *World)
		decode_member seq index env
		# (name, valuePtr, env) = corba_nv_pair_seq_get seq index env
		# (value, env)			= corba_any_decode valuePtr env
		= ((name, value), env)

corba_any_seq_encode :: [Any] *World -> (AnySeq, *World)
corba_any_seq_encode members env
	= corba_seq_encode (corba_any_seq_create) (add_member) members env
	where
		add_member :: AnySeq Int Any *World -> *World
		add_member seq index value env
		# (valuePtr, env)		= corba_any_encode value env
		# env 					= corba_any_seq_fill seq index valuePtr env
		# env					= corba_any_delete valuePtr env
		= env

corba_any_seq_decode :: AnySeq *World 
							-> ([Any], *World)
corba_any_seq_decode seq env
	# (len, env) = corba_any_seq_length seq env
	= corba_seq_decode decode_member seq len env
	where
		decode_member :: AnySeq Int *World -> (Any, *World)
		decode_member seq index env
		# (valuePtr, env) 	= corba_any_seq_get seq index env
		# (value, env)		= corba_any_decode valuePtr env
		= (value, env)

corba_any_encode :: Any *World -> (AnyPtr, *World)
corba_any_encode any env = create any
where
	create (CorbaNull) 			= corba_any_create_null env
	create (CorbaVoid) 			= 
				abort "corba_any_create called with CorbaVoid argument"
	create (CorbaShort s) 		= corba_any_create_short s env
	create (CorbaUShort s) 		= corba_any_create_ushort s env
	create (CorbaLong s) 		= corba_any_create_long s env
	create (CorbaULong s) 		= corba_any_create_ulong s env
	create (CorbaFloat f) 		= corba_any_create_float f env
	create (CorbaDouble d) 		= corba_any_create_double d env
	create (CorbaBoolean b) 	= corba_any_create_bool b env
	create (CorbaChar c) 		= corba_any_create_char c env
	create (CorbaOctet s)		= corba_any_create_octet s env
	create (CorbaString s)		= corba_any_create_string s env

	create (CorbaAny any)
	# (anyPtr, env)	=	corba_any_encode any env
	# (res, env) = corba_any_create_any anyPtr env
	= (res, env)

	create (CorbaTypeCode tc)
	# (tcPtr, env)	=	corba_tc_encode tc env
	= corba_any_create_tc tcPtr env

	create (CorbaObjRef obj)
	= corba_any_create_objref obj.CORBA_Object.reference env

	create (CorbaStruct tc members)
	# (tcPtr, env)		=	corba_tc_encode tc env
	# (membersPtr, env)	=	corba_nv_pair_seq_encode members env
	# (res, env)		=	corba_dyn_struct_create tcPtr membersPtr env
	= (res, env)

	create (CorbaSequence tc elements)
	# (tcPtr, env)			=	corba_tc_encode tc env
	# (elementsPtr, env)	= 	corba_any_seq_encode elements env
	# (res, env)			=	corba_dyn_sequence_create tcPtr elementsPtr env
	= (res, env)

	create (CorbaUnion tc discriminant value)
	# (tcPtr, env)				=	corba_tc_encode tc env
	# (discriminantPtr, env)	=	corba_any_encode discriminant env
	# (valuePtr, env)			=	corba_any_encode value env
	= corba_dyn_union_create tcPtr discriminantPtr valuePtr env

	create (CorbaEnum tc value)
	# (tcPtr, env)				=	corba_tc_encode tc env
	= corba_dyn_enum_create tcPtr value env

	create (CorbaArray tc elements)
	# (tcPtr, env)				=	corba_tc_encode tc env
	# (elementsPtr, env)		=	corba_any_seq_encode elements env
	= corba_dyn_array_create tcPtr elementsPtr env

	// Same as for CorbaStructs...
	create (CorbaException tc members)
	# (tcPtr, env)		=	corba_tc_encode tc env
	# (membersPtr, env)	=	corba_nv_pair_seq_encode members env
	# (res, env)		=	corba_dyn_struct_create tcPtr membersPtr env
	= (res, env)

corba_any_decode :: AnyPtr *World -> (Any, *World)
corba_any_decode any env
	# (tcPtr, env)	=	corba_any_get_type any env
	# (tc, env)		=	corba_tc_decode tcPtr env
	= decode any tc env
where
	decode any (Tk_Null) env  = (CorbaNull, env)
	decode any (Tk_Void) env  = (CorbaVoid, env)

	decode any (Tk_Short) env 
	# (value, env)		=	corba_any_extract_short any env
	= (CorbaShort value, env)

	decode any (Tk_Long) env 
	# (value, env)		=	corba_any_extract_long any env
	= (CorbaLong value, env)

	decode any (Tk_UShort) env 
	# (value, env)		=	corba_any_extract_ushort any env
	= (CorbaUShort value, env)

	decode any (Tk_ULong) env
	# (value, env)		=	corba_any_extract_ulong any env
	= (CorbaULong value, env)

	decode any (Tk_Float) env
	# (value, env)		=	corba_any_extract_float any env
	= (CorbaFloat value, env)

	decode any (Tk_Double) env 
	# (value, env)		=	corba_any_extract_double any env
	= (CorbaDouble value, env)

	decode any (Tk_Boolean) env
	# (value, env)		=	corba_any_extract_boolean any env
	= (CorbaBoolean value, env)

	decode any (Tk_Char) env
	# (value, env)		=	corba_any_extract_char any env
	= (CorbaChar value, env)

	decode any (Tk_Octet) env
	# (value, env)		=	corba_any_extract_octet any env
	= (CorbaOctet value, env)

	decode any (Tk_String l) env
	# (value, env)		=	corba_any_extract_string any env
	= (CorbaString value, env)

	decode any (Tk_Any) env
	# (anyPtr, env)		=	corba_any_extract_any any env
	# (value, env)		=	corba_any_decode anyPtr env
	= (CorbaAny value, env)

	decode any (Tk_TypeCode) env
	# (tcPtr, env)			=	corba_any_extract_tc any env
	# (value, env)				=	corba_tc_decode tcPtr env
	= (CorbaTypeCode value, env)

	decode any (Tk_ObjRef repoid name) env 
	# (ref, env)			=	corba_any_extract_objref any env
	= (CorbaObjRef { CORBA_Object | reference = ref}, env)

	decode any (Tk_Struct repoid name members) env
	# (seqPtr, env) 			= corba_dyn_struct_members any env
	# (seq, env)				= corba_nv_pair_seq_decode seqPtr env
	# env						= corba_nv_pair_seq_delete seqPtr env
	= (CorbaStruct (Tk_Struct repoid name members) seq, env)

	decode any (Tk_Exception repoid name members) env
	# (seqPtr, env) 			= corba_dyn_struct_members any env
	# (seq, env)				= corba_nv_pair_seq_decode seqPtr env
	# env						= corba_nv_pair_seq_delete seqPtr env
	= (CorbaException (Tk_Exception repoid name members) seq, env)

	decode any (Tk_Sequence bound tc) env
	# (seqPtr, env) 			= corba_dyn_sequence_elements any env
	# (seq, env)				= corba_any_seq_decode seqPtr env
	# env						= corba_any_seq_delete seqPtr env
	= (CorbaSequence (Tk_Sequence bound tc) seq, env)

	decode any (Tk_Union repoid name tc members) env
	# (labelPtr, valuePtr, env) 	= corba_dyn_union_get any env
	# (label, env)					= corba_any_decode labelPtr env
	# (value, env)					= corba_any_decode valuePtr env
	= (CorbaUnion (Tk_Union repoid name tc members) label value, env)

	decode any (Tk_Enum repoid name members) env
	# (value, env) = corba_dyn_enum_get any env
	= (CorbaEnum (Tk_Enum repoid name members) value, env)

	decode any (Tk_Array bound tc) env
	# (seqPtr, env) 				= corba_dyn_array_elements any env
	# (seq, env)					= corba_any_seq_decode seqPtr env
	# env							= corba_any_seq_delete seqPtr env
	= (CorbaArray (Tk_Array bound tc) seq, env)

	decode any (Tk_Alias _ _ tc) env = decode any tc env

	decode any tc env = abort ("corba_any_decode called with " +++ (toStr tc) +++ " argument.")
	
////////////////////////////////
//        EXCEPTION
////////////////////////////////

// HIGHER LEVEL INTERFACE

:: CompletionStatus = Completed_Yes | Completed_No | Completed_Maybe
:: CorbaException = NoException | 
					SystemException String CompletionStatus Int | 
					UserException String Any

corba_exception_decode :: ExceptionPtr [TypeCode] *World -> (CorbaException, *World)
corba_exception_decode ex exception_defs env
	# (is_system, env) = corba_exception_is_system ex env
	# (repoid, env) = corba_exception_get_repoid ex env
	| is_system == True
		# (minor, env) = corba_exception_get_minor ex env
		# (completed, env) = corba_exception_get_completed ex env
	= (SystemException repoid (decode_completed completed) minor, env)
	| otherwise
	= decode_user_exception ex repoid env
	where
		decode_user_exception ex repoid env
		# (unknown, env) = corba_unknown_ex_narrow ex env
		| unknown <> 0
		# (repoid, env) 	= corba_unknown_ex_get_repoid unknown env
		= decode_unknown unknown repoid env
		| otherwise
		= (UserException repoid (CorbaNull), env)
		where
			// Find the exception definition corresponding to 'repoid',
			// and try to decode the contents of the exception according to
			// it.
			decode_unknown unknown repoid env
			# defs = filter find_def exception_defs
			| isEmpty defs
			= (UserException repoid (CorbaNull), env)
			| otherwise
			# (tcPtr, env)			= corba_tc_encode (hd defs) env
			# (valuePtr, env) 	= corba_unknown_ex_get_value unknown tcPtr env
			# (value, env)			= corba_any_decode valuePtr env
			= (UserException repoid value, env)
			where
				find_def (Tk_Exception repoid2 _ _) = (repoid == repoid2)
				find_def any = abort ("find_def called with non exception argument: " +++ (toStr any))

		decode_completed :: Int -> CompletionStatus
		decode_completed COMPLETED_YES 		= Completed_Yes
		decode_completed COMPLETED_NO 		= Completed_No
		decode_completed COMPLETED_MAYBE 	= Completed_Maybe
		decode_completed _					= abort "bad value for CompletionStatus."

corba_exception_encode :: CorbaException -> Any
corba_exception_encode (UserException repoid ex) = ex

////////////////////////////////
//         NAMEDVALUE
////////////////////////////////

:: CorbaArg = 	InArg String Any 
			| 	OutArg String Any 
			| 	InOutArg String Any
			| 	InArgTemplate String TypeCode
			| 	OutArgTemplate String TypeCode
			| 	InOutArgTemplate String TypeCode

corba_arg_get_value :: CorbaArg -> Any
corba_arg_get_value (InArg _ v) 	= v
corba_arg_get_value (OutArg _ v) 	= v
corba_arg_get_value (InOutArg _ v) 	= v
corba_arg_get_value arg				= abort ("corba_arg_get_value called with " +++ (toStr arg) +++ " argument.")

corba_nv_decode :: NamedValuePtr *World -> (CorbaArg, *World)
corba_nv_decode nv env
	# (name, env) 		= corba_nv_get_name nv env
	# (valuePtr, env) 	= corba_nv_get_value nv env
	# (flags, env) 		= corba_nv_get_flags nv env
	# (value, env) 		= corba_any_decode valuePtr env
	= (decode name value flags, env)
	where
		decode :: String Any Int -> CorbaArg
		decode name value ARG_IN 	= InArg name value
		decode name value ARG_OUT 	= OutArg name value
		decode name value ARG_INOUT = InOutArg name value
		decode name value _		 	= abort "bad flags."

corba_nvlist_decode :: !NVListPtr *World -> ([CorbaArg], *World)
corba_nvlist_decode nvlist env
	# (len, env) = corba_nvlist_count nvlist env
	= corba_seq_decode decode_member nvlist len env
	where
		decode_member nvlist index env
		# (nvPtr, env) 			= corba_nvlist_get nvlist index env
		# (nv, env)				= corba_nv_decode nvPtr env
		= (nv, env)

corba_nvlist_encode :: [CorbaArg] *World -> (NVListPtr, *World)
corba_nvlist_encode members env
	= corba_seq_encode (\i -> corba_nvlist_create) (add_member) members env
	where
		add_member :: NVListPtr Int CorbaArg *World -> *World

		add_member seq index (InArg name value) env
			# (nv, env)			= corba_nvlist_add seq ARG_IN env
			# (valuePtr, env)	= corba_any_encode value env
			# env				= corba_nv_set_value nv valuePtr env
			= env

		add_member seq index (OutArg name value) env
			# (nv, env)			= corba_nvlist_add seq ARG_OUT env
			# (valuePtr, env)	= corba_any_encode value env
			# env				= corba_nv_set_value nv valuePtr env
			= env

		add_member seq index (InOutArg name value) env
			# (nv, env)			= corba_nvlist_add seq ARG_INOUT env
			# (valuePtr, env)	= corba_any_encode value env
			# env				= corba_nv_set_value nv valuePtr env
			= env

		add_member seq index (InArgTemplate name tc) env
			# (nv, env)			= corba_nvlist_add seq ARG_IN env
			# (tcPtr, env)		= corba_tc_encode tc env
			# env				= corba_nv_set_type nv tcPtr env
			= env

		add_member seq index (OutArgTemplate name tc) env
			# (nv, env)			= corba_nvlist_add seq ARG_OUT env
			# (tcPtr, env)		= corba_tc_encode tc env
			# env				= corba_nv_set_type nv tcPtr env
			= env

		add_member seq index (InOutArgTemplate name tc) env
			# (nv, env)			= corba_nvlist_add seq ARG_INOUT env
			# (tcPtr, env)		= corba_tc_encode tc env
			# env				= corba_nv_set_type nv tcPtr env
			= env

////////////////////////////////
//          REQUEST
////////////////////////////////

corba_invoke :: CORBA_Object String [CorbaArg] TypeCode [TypeCode] *World -> (Any, CorbaException, [CorbaArg], *World)
corba_invoke obj operation args result_tc raises env
	# obj = obj.CORBA_Object.reference
	# (req, env) = corba_object_create_request obj operation env
	# (resultTcPtr, env) = corba_tc_encode result_tc env
	# env = corba_request_set_result_tc req resultTcPtr env
	# env = foldl (\e v -> add_arg req v e) env args
	# env = foldl (\e v -> add_exception req v e) env raises
	# env = corba_request_invoke req env
	# (ex, env) = corba_request_exception req env
	| ex <> 0
		# (ex, env)		= corba_exception_decode ex raises env
	= (CorbaVoid, ex, [], env) 
	# (resPtr, env) = corba_request_result req env
	# (params, env) = get_out_params req env
	# (res, env)	= corba_any_decode resPtr env
	# env			= corba_request_delete req env
	= (res, NoException, params, env)
	where
		// Add arguments to the request
		add_arg :: RequestPtr CorbaArg *World -> *World
		add_arg req (InArg name value) env
		# (valuePtr, env)		=	corba_any_encode value env
	 	# env = corba_request_add_in_arg req name valuePtr env
		# env = corba_any_delete valuePtr env
		= env

		add_arg req (OutArg name value) env
		# (valuePtr, env)		=	corba_any_encode value env		
	 	# env = corba_request_add_out_arg req name valuePtr env
		# env = corba_any_delete valuePtr env
		= env

		add_arg req (InOutArg name value) env
		# (valuePtr, env)		=	corba_any_encode value env		
	 	# env = corba_request_add_inout_arg req name valuePtr env
		# env = corba_any_delete valuePtr env
		= env

		add_arg req (OutArgTemplate name tc) env
		# (tcPtr, env)			=	corba_tc_encode tc env
		= corba_request_add_out_arg_template req name tcPtr env

		// Add a possible exception
		add_exception :: RequestPtr TypeCode *World -> *World
		add_exception req tc env
		# (tcPtr, env)		=	corba_tc_encode tc env
		 = corba_request_add_exception_def req tcPtr env

		// Collect the out and inout arguments
		get_out_params req env
		# (nvlistPtr, env) 		= corba_request_arguments req env
		# (nvlist, env)			= corba_nvlist_decode nvlistPtr env
		= (filter is_out_param nvlist, env)

		is_out_param :: CorbaArg -> Bool
		is_out_param (InArg _ _) = False
		is_out_param _			= True


//////////////////////////////////////////////////////////////////////
//             DEFINITIONS USED BY THE GENERATED CODE
//////////////////////////////////////////////////////////////////////

////////////////////////////////
//        CORBA::Void
////////////////////////////////

// There is no void type in Clean
:: CORBA_Void	:==	Void

CORBA_Void__tc	::	TypeCode
CORBA_Void__tc	=	Tk_Void

CORBA_Void__decode :: Any *World	->	(MayBe CORBA_Void, *World)
CORBA_Void__decode (CorbaVoid)	env =	(Just Void, env)
CORBA_Void__decode _ env			=	(Nothing, env)

CORBA_Void__encode :: CORBA_Void -> Any
CORBA_Void__encode _ = (CorbaVoid)

////////////////////////////////
//        CORBA::Short
////////////////////////////////

:: CORBA_Short	:==	Int

CORBA_Short__tc	::	TypeCode
CORBA_Short__tc	=	Tk_Short

CORBA_Short__decode :: Any *World	->	(MayBe CORBA_Short, *World)
CORBA_Short__decode (CorbaShort i)	env	=	(Just i, env)
CORBA_Short__decode _ env				=	(Nothing, env)

CORBA_Short__encode :: CORBA_Short -> Any
CORBA_Short__encode i = (CorbaShort i)

////////////////////////////////
//        CORBA::Long
////////////////////////////////

:: CORBA_Long	:==	Int

CORBA_Long__tc	::	TypeCode
CORBA_Long__tc	=	Tk_Long

CORBA_Long__decode :: Any *World	->	(MayBe CORBA_Long, *World)
CORBA_Long__decode (CorbaLong i) env	=	(Just i, env)
CORBA_Long__decode _ env				=	(Nothing, env)

CORBA_Long__encode :: CORBA_Long -> Any
CORBA_Long__encode i = (CorbaLong i)

////////////////////////////////
//        CORBA::UShort
////////////////////////////////

:: CORBA_UShort	:==	Int

CORBA_UShort__tc	::	TypeCode
CORBA_UShort__tc	=	Tk_UShort

CORBA_UShort__decode :: Any *World	->	(MayBe CORBA_UShort, *World)
CORBA_UShort__decode (CorbaUShort i) env	=	(Just i, env)
CORBA_UShort__decode _	env					=	(Nothing, env)

CORBA_UShort__encode :: CORBA_UShort -> Any
CORBA_UShort__encode i = (CorbaUShort i)

////////////////////////////////
//        CORBA::ULong
////////////////////////////////

:: CORBA_ULong	:==	Int

CORBA_ULong__tc	::	TypeCode
CORBA_ULong__tc	=	Tk_ULong

CORBA_ULong__decode :: Any *World	->	(MayBe CORBA_ULong, *World)
CORBA_ULong__decode (CorbaULong i) env	=	(Just i, env)
CORBA_ULong__decode _ env				=	(Nothing, env)

CORBA_ULong__encode :: CORBA_ULong -> Any
CORBA_ULong__encode i = (CorbaULong i)

////////////////////////////////
//        CORBA::Float
////////////////////////////////

:: CORBA_Float	:==	Real

CORBA_Float__tc	::	TypeCode
CORBA_Float__tc	=	Tk_Float

CORBA_Float__decode :: Any *World	->	(MayBe CORBA_Float, *World)
CORBA_Float__decode (CorbaFloat f) env	=	(Just f, env)
CORBA_Float__decode _ env				=	(Nothing, env)

CORBA_Float__encode :: CORBA_Float -> Any
CORBA_Float__encode f = (CorbaFloat f)

////////////////////////////////
//        CORBA::Double
////////////////////////////////

:: CORBA_Double	:==	Real

CORBA_Double__tc	::	TypeCode
CORBA_Double__tc	=	Tk_Double

CORBA_Double__decode :: Any *World	->	(MayBe CORBA_Double, *World)
CORBA_Double__decode (CorbaDouble f) env	=	(Just f, env)
CORBA_Double__decode _ env					=	(Nothing, env)

CORBA_Double__encode :: CORBA_Double -> Any
CORBA_Double__encode f = (CorbaDouble f)

////////////////////////////////
//        CORBA::Boolean
////////////////////////////////

:: CORBA_Boolean	:==	Bool

CORBA_Boolean__tc	::	TypeCode
CORBA_Boolean__tc	=	Tk_Boolean

CORBA_Boolean__decode :: Any *World	->	(MayBe CORBA_Boolean, *World)
CORBA_Boolean__decode (CorbaBoolean b) env	=	(Just b, env)
CORBA_Boolean__decode _ env				=	(Nothing, env)

CORBA_Boolean__encode :: CORBA_Boolean -> Any
CORBA_Boolean__encode b = (CorbaBoolean b)

////////////////////////////////
//        CORBA::Char
////////////////////////////////

:: CORBA_Char	:==	Char

CORBA_Char__tc	::	TypeCode
CORBA_Char__tc	=	Tk_Char

CORBA_Char__decode :: Any *World	->	(MayBe CORBA_Char, *World)
CORBA_Char__decode	(CorbaChar c) env	=	(Just c, env)
CORBA_Char__decode	_ env				=	(Nothing, env)

CORBA_Char__encode :: CORBA_Char -> Any
CORBA_Char__encode c = (CorbaChar c)

////////////////////////////////
//        CORBA::Octet
////////////////////////////////

:: CORBA_Octet	:==	Int

CORBA_Octet__tc	::	TypeCode
CORBA_Octet__tc	=	Tk_Octet

CORBA_Octet__decode :: Any *World	->	(MayBe CORBA_Octet, *World)
CORBA_Octet__decode (CorbaOctet i) env	=	(Just i, env)
CORBA_Octet__decode _ env				=	(Nothing, env)

CORBA_Octet__encode :: CORBA_Octet -> Any
CORBA_Octet__encode i = (CorbaOctet i)

////////////////////////////////
//        CORBA::Any
////////////////////////////////

:: CORBA_Any	:==	Any

CORBA_Any__tc	::	TypeCode
CORBA_Any__tc	=	Tk_Any

CORBA_Any__decode :: Any *World	->	(MayBe CORBA_Any, *World)
CORBA_Any__decode (CorbaAny any) env	=	(Just any, env)
CORBA_Any__decode _ env				=	(Nothing, env)

CORBA_Any__encode :: Any -> Any
CORBA_Any__encode any = (CorbaAny any)

////////////////////////////////
//     CORBA::TypeCode
////////////////////////////////

:: CORBA_TypeCode	:==	TypeCode

CORBA_TypeCode__tc :: TypeCode
CORBA_TypeCode__tc	=	Tk_TypeCode

CORBA_TypeCode__decode :: Any *World -> (MayBe CORBA_TypeCode, *World)
CORBA_TypeCode__decode (CorbaTypeCode tc) env	= (Just tc, env)
CORBA_TypeCode__decode _ env					= (Nothing, env)

CORBA_TypeCode__encode :: TypeCode -> Any
CORBA_TypeCode__encode tc = (CorbaTypeCode tc)

////////////////////////////////
//      CORBA::String
////////////////////////////////

// This is the unbounded version
:: CORBA_String	:==	String

CORBA_String__tc :: TypeCode
CORBA_String__tc = Tk_String 0

CORBA_String__decode :: Any *World -> (MayBe CORBA_String, *World)
CORBA_String__decode (CorbaString s) env 	= (Just s, env)
CORBA_String__decode _ env					= (Nothing, env)

CORBA_String__encode :: CORBA_String -> Any
CORBA_String__encode s = (CorbaString s)

////////////////////////////////
//      CORBA::Object
////////////////////////////////

:: CORBA_Object  = { reference :: CorbaObjectPtr }

CORBA_Object__create :: CORBA_Object -> CORBA_Object
CORBA_Object__create obj = obj

CORBA_Object__tc :: TypeCode
CORBA_Object__tc = (Tk_ObjRef "IDL:omg.org/CORBA/Object:1.0" "Object")

CORBA_Object__decode :: Any *World -> (MayBe CORBA_Object, *World)
CORBA_Object__decode (CorbaObjRef ref) env = (Just ref, env)
CORBA_Object__decode _ env					= (Nothing, env)

CORBA_Object__encode :: CORBA_Object -> Any
CORBA_Object__encode obj = (CorbaObjRef obj)

CORBA_Object_is_a :: CORBA_Object String *World -> (Bool, *World)
CORBA_Object_is_a obj repoid env = corba_object_is_a obj.CORBA_Object.reference repoid env

CORBA_Object__nil :: *World -> (CORBA_Object, *World)
CORBA_Object__nil env
	# (ref, env)		=	corba_object_nil env
	= ({ CORBA_Object | reference = ref }, env)

class CORBA_Object__widen_class a where
	CORBA_Object__widen :: a -> CORBA_Object

instance CORBA_Object__widen_class CORBA_Object where
	CORBA_Object__widen obj = obj

////////////////////////////////
//         ORB
////////////////////////////////

:: CORBA_ORB :== ORBPtr

CORBA_ORB_init :: [String] *World -> (CORBA_ORB, [String], *World)
CORBA_ORB_init args env
	# (argv, argc, env)					= c_string_array_encode args env
	# (orb, out_argv, out_argc, env)	= corba_orb_init (argv, argc) env
	# (out_args, env)					= c_string_array_decode out_argv out_argc env
	= (orb, out_args, env)

CORBA_ORB_bind :: CORBA_ORB String String *World -> (MayBe CORBA_Object, *World)
CORBA_ORB_bind orb address repoid env
	# (obj, env) = corba_orb_bind orb address repoid env
	| obj == 0
	= (Nothing, env)
	| otherwise
	= (Just ({ CORBA_Object | reference = obj }), env)

CORBA_ORB_string_to_object :: CORBA_ORB String *World -> (MayBe CORBA_Object, *World)
CORBA_ORB_string_to_object orb ior env
	# (obj, env) = corba_orb_string_to_object orb ior env
	| obj == 0
	= (Nothing, env)
	| otherwise
	= (Just ({ CORBA_Object | reference = obj }), env)

CORBA_ORB_object_to_string :: CORBA_ORB CORBA_Object *World -> (String, *World)
CORBA_ORB_object_to_string orb obj env
	= corba_orb_object_to_string orb obj.CORBA_Object.reference env

CORBA_ORB_resolve_initial_references :: CORBA_ORB String *World -> (MayBe CORBA_Object, *World)
CORBA_ORB_resolve_initial_references orb name env
	# (obj, env) = corba_orb_resolve_initial_references orb name env
	| obj == 0
	= (Nothing, env)
	| otherwise
	= (Just ({ CORBA_Object | reference = obj }), env)

WriteIORToFile :: CORBA_ORB CORBA_Object String *World -> *World
WriteIORToFile orb obj filename env
	# (ior, env) 		= CORBA_ORB_object_to_string orb obj env
	# (success, f, env)	= fopen filename FWriteText env
	| not success
	= env
	| otherwise
	# f					= f <<< ior
	# (_, env)			= fclose f env
	= env

ReadIORFromFile :: CORBA_ORB String *World -> (MayBe CORBA_Object, *World)
ReadIORFromFile orb filename env
	# (success, f, env)	= fopen filename FReadText env
	| not success
	= (Nothing, env)
	| otherwise
	# (ior,f)			= freadline f
	# (_, env)			= fclose f env
	= CORBA_ORB_string_to_object orb ior env

////////////////////////////////
//      SERVER REQUEST
////////////////////////////////

:: CORBA_ServerRequest :== ServerRequestPtr

CORBA_ServerRequest_operation :: CORBA_ServerRequest *env -> (String, *env)
CORBA_ServerRequest_operation req w = corba_server_request_operation req w

CORBA_ServerRequest_set_arg_types :: CORBA_ServerRequest [CorbaArg] *World -> *World
CORBA_ServerRequest_set_arg_types req args w
	# (nvlist, w) 	= corba_nvlist_encode args w
	# w				= corba_server_request_set_params req nvlist w
	= w

CORBA_ServerRequest_get_arg :: CORBA_ServerRequest Int *World -> (Any, *World)
CORBA_ServerRequest_get_arg req index w
	# (nvlist, w)	= corba_server_request_get_args req w
	# (nv, w)		= corba_nvlist_get nvlist index w
	# (anyPtr, w)	= corba_nv_get_value nv w
	# (any, w)		= corba_any_decode anyPtr w
	= (any, w)

CORBA_ServerRequest_set_arg :: CORBA_ServerRequest Int Any *World -> *World
CORBA_ServerRequest_set_arg req index value w
	# (nvlist, w)	= corba_server_request_get_args req w
	# (nv, w)		= corba_nvlist_get nvlist index w
	# (anyPtr, w)	= corba_any_encode value w
	# w				= corba_nv_set_value nv anyPtr w
	= w

CORBA_ServerRequest_set_result :: CORBA_ServerRequest Any *World -> *World
CORBA_ServerRequest_set_result req result w
	# (anyPtr, w)	= corba_any_encode result w
	= corba_server_request_set_result req anyPtr w


CORBA_ServerRequest_set_exception :: CORBA_ServerRequest Any *World -> *World
CORBA_ServerRequest_set_exception req ex w
	# (anyPtr, w)	= corba_any_encode ex w
	= corba_server_request_set_exception req anyPtr w

////////////////////////////////
//        SERVER SIDE
////////////////////////////////

:: PSt l
	= {	ls :: l
	,	io :: IOSt l
	}


:: IOSt l = {
	servants 		:: [(CorbaObjectPtr, CORBA_Servant_Wrapper l World)],
	need_to_exit 	:: Bool,
	current			:: CORBA_Object,	// The currently executing object
	orb				:: CORBA_ORB
	};

:: ServerInitFunc l *env :== (PSt l) env -> ((PSt l), env)

NoInit :: (PSt l) *env -> ((PSt l), *env)
NoInit ps env = (ps, env)

// Empty local state
:: Void = Void

:: CORBA_Servant_Wrapper l *env = E.obj: {
	obj 		:: obj,
	dispatch 	:: (PSt l) CORBA_ServerRequest obj env -> (((PSt l), obj), env)
	};

CORBA_Server_run :: CORBA_ORB l (ServerInitFunc l *World) *World -> *World
CORBA_Server_run orb initial_state init_func w
	# ps 		= initial_ps
	# w			= corba_listener_init w
	# (ps, w) 	= init_func ps w
	# (_, w) 	= event_loop ps w
	= w
where
	initial_ps = {
		ls = initial_state,
		io = {
			servants 		= [],
			need_to_exit	= False,
			current			= { CORBA_Object | reference = 0 },
			orb				= orb
		}
		}

	event_loop :: (PSt l) *World -> ((PSt l), *World)
	event_loop ps=:{io} w
		# (req, obj, w) = corba_listener_receive w
		# (serv, index)	= lookup_servant io.servants obj 0
		# ps			= {ps & io.current = { CORBA_Object | reference = obj }}
		# (ps, serv, w)	= dispatch_to ps req serv w
		// This may be a bit slow
		# ps			= {ps & io.servants = updateAt index (obj, serv) io.servants}
		| ps.io.need_to_exit
		= (ps, w)
		| otherwise
		= event_loop ps w
	where
		dispatch_to :: (PSt l) ServerRequestPtr (CORBA_Servant_Wrapper l *World) *World -> ((PSt l), (CORBA_Servant_Wrapper l *World), *World)
		dispatch_to ps req serv=:{obj, dispatch} w
			# ((ps, obj), w) = dispatch ps req obj w
			# w				 = corba_listener_reply w
			= (ps, {serv & obj=obj}, w)

		// The ORB takes care of raising errors for nonexistent servants
		lookup_servant [(o, s):xs] obj index
			| o == obj
			= (s, index)
			| otherwise
			= lookup_servant xs obj (index + 1)
		lookup_servant [] obj index = abort "Nonexistent servant."


// Instruct the CORBA server process to exit
CORBA_Server_exit :: (PSt l) -> (PSt l)
CORBA_Server_exit ps=:{io} = {ps & io.need_to_exit = True}

// Registers a new CORBA servant
CORBA_Server_add_servant :: (PSt l) String (CORBA_Servant_Wrapper l *World) *World -> (CORBA_Object, (PSt l), *World)
CORBA_Server_add_servant ps=:{io} repoid servant w
	# (obj, w)			= corba_servant_create repoid w
	= (	{ CORBA_Object | reference = obj}, 
		{ps & io.servants = [(obj, servant):io.servants]}, w)

CORBA_Server_get_current :: (PSt l) -> CORBA_Object
CORBA_Server_get_current pst=:{io} = io.IOSt.current

CORBA_Server_get_orb :: (PSt l) -> CORBA_ORB
CORBA_Server_get_orb pst=:{io} = io.IOSt.orb

// Local Variables:
// tab-width: 4
// End:



